/**
 * OW2 Util
 * Copyright (C) 2008 SERLI
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id epositoryDeployerImpl.java $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.plan.deployer.impl;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.plan.bindings.repository.Repositories;
import org.ow2.util.plan.bindings.repository.Repository;
import org.ow2.util.plan.deploy.deployable.api.RepositoryDeployable;
import org.ow2.util.plan.deployer.api.IRepositoryDeployer;
import org.ow2.util.plan.reader.repository.IRepositoryDataReader;
import org.ow2.util.plan.repository.api.IRepositoryManager;

/**
 * Implementation for the deployer for repository deployables.
 * @author Mickaël LEDUQUE
 */
public abstract class AbsRepositoryDeployerImpl implements IRepositoryDeployer {

    /**
     * The logger.
     */
    private Log logger = LogFactory.getLog(AbsRepositoryDeployerImpl.class);

    /**
     * The deployed repository deployables.
     */
    private Collection<RepositoryDeployable> repositoryDeployables = new ArrayList<RepositoryDeployable>();

    /**
     * Repository reader.
     */
    private IRepositoryDataReader repositoryReader = null;

    /**
     * {@inheritDoc}
     */
    public void deploy(final IDeployable<?> deployable) throws DeployerException {
        if (deployable == null) {
            throw new DeployerException("Null deployable");
        }
        if (!(deployable instanceof RepositoryDeployable)) {
            throw new DeployerException("Bad deployable class: " + deployable.getClass());
        }
        RepositoryDeployable repositoryDeployable = (RepositoryDeployable) deployable;

        Repositories repositories = null;
        if (repositoryReader == null) {
            logger.error("Deployment aborted - the RepositoryDataReader is null");
            throw new DeployerException("the RepositoryDataReader is null");
        }
        URL archiveUrl = null;
        try {
            archiveUrl = repositoryDeployable.getArchive().getURL();
        } catch (ArchiveException e) {
            logger.error("Error while getting URL for archive {0} - Deployment aborted", repositoryDeployable.getArchive());
            throw new DeployerException(e);
        }
        try {
            repositories = repositoryReader.readRepositories(new File(archiveUrl.getPath()));
        } catch (Exception e) {
            logger.error("Error while parsing archive {0} - Deployment aborted", archiveUrl);
            throw new DeployerException(e);
        }

        if (repositories == null) {
            logger.error("The deployable {0} resolved in null repository data - Deployment aborted", deployable);
            throw new DeployerException("null repositories structure");
        }

        repositoryDeployable.setAttachedData(repositories);
        this.addToRepositoryManager(repositories); // may fail with DeployerException

        this.repositoryDeployables.add(repositoryDeployable);
        logger.debug("Repository Deployable {0} deployed", deployable);
        StringBuilder repositorySB = new StringBuilder();
        for (Repository repository : repositories.getRepositories()) {
            repositorySB.append("\t- ");
            repositorySB.append(repository.toString());
            repositorySB.append("\n");
        }
        logger.debug("- {0} repositories added :\n{1}", repositories.getRepositories().size(), repositorySB.toString());
    }

    /**
     * Add the repositories to the repository manager.
     * @param repositories the added repositories.
     * @throws DeployerException if the addition failed.
     */
    protected void addToRepositoryManager(final Repositories repositories) throws DeployerException {
        IRepositoryManager repositoryManager = this.getRepositoryManager();
        for (Repository repository : repositories.getRepositories()) {
            try {
                repositoryManager.addRepository(repository);
            } catch (Exception e) {
                /* Clean up the repository manager before failing */
                for (Repository deletedRepository : repositories.getRepositories()) {
                    repositoryManager.removeRepository(deletedRepository);
                }
                throw new DeployerException(e);
            }
        }
    }

    /**
     * Remove undeployed repository data from the repositories manager.
     * @param repositories the removed repository data.
     */
    protected void removeFromRepositoryManager(final Repositories repositories) {
        IRepositoryManager repositoryManager = this.getRepositoryManager();
        for (Repository repository : repositories.getRepositories()) {
            repositoryManager.removeRepository(repository);
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isDeployed(final IDeployable<?> deployable) throws DeployerException {
        if (deployable == null) {
            throw new DeployerException("Null deployable");
        }
        if (!(deployable instanceof RepositoryDeployable)) {
            throw new DeployerException("Bad deployable class: " + deployable.getClass());
        }

        return this.repositoryDeployables.contains(deployable);
    }

    /**
     * {@inheritDoc}
     */
    public boolean supports(final IDeployable<?> deployable) {
        return (deployable instanceof RepositoryDeployable);
    }

    /**
     * {@inheritDoc}
     */
    public void undeploy(final IDeployable<?> deployable) throws DeployerException {
        if (deployable == null) {
            throw new DeployerException("Null deployable");
        }
        if (!this.supports(deployable)) {
            throw new DeployerException("Bad deployable class: " + deployable.getClass());
        }
        RepositoryDeployable repositoryDeployable = (RepositoryDeployable) deployable;
        // Check if the deployable is deployed
        if (!this.repositoryDeployables.contains(repositoryDeployable)) {
            throw new DeployerException("Cannot undeploy repository deployable " + deployable.toString()
                    + ", it is not deployed");

        }
        Repositories repositories = repositoryDeployable.getAttachedData();

        this.removeFromRepositoryManager(repositories);

        repositoryDeployables.remove(repositoryDeployable);
        logger.info("Repository deployable {0} undeployed", deployable);
    }

    /**
     * {@inheritDoc}
     */
    public void setReader(final IRepositoryDataReader reader) {
        this.repositoryReader = reader;
    }

    /**
     * {@inheritDoc}
     */
    public IRepositoryDataReader getReader() {
        return this.repositoryReader;
    }

    /**
     * {@inheritDoc}
     */
    public int getDeployablesCount() {
        return this.repositoryDeployables.size();
    }

    /**
     * Returns the Repository Manager that is used by the deployer.
     * @return the Repository Manager that is used by the deployer
     */
    protected abstract IRepositoryManager getRepositoryManager();
}
