/**
 * OW2 Util
 * Copyright (C) 2008
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: DeploymentPlanDeployer.java $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.plan.deployer.impl;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.OSGiDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.ee.deploy.api.deployer.IDeployerManager;
import org.ow2.util.ee.deploy.api.deployer.UnsupportedDeployerException;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.plan.bindings.deploymentplan.Deployment;
import org.ow2.util.plan.bindings.deploymentplan.DeploymentHelper;
import org.ow2.util.plan.bindings.deploymentplan.DeploymentPlan;
import org.ow2.util.plan.bindings.deploymentplan.DeploymentPlanFragment;
import org.ow2.util.plan.bindings.exceptions.InvalidDeploymentException;
import org.ow2.util.plan.deploy.deployable.api.DeploymentPlanDeployable;
import org.ow2.util.plan.deploy.deployable.impl.DeploymentPlanDeployableImpl;
import org.ow2.util.plan.deployer.api.DeploymentPlanDeploymentException;
import org.ow2.util.plan.deployer.api.DeploymentPlanUndeploymentException;
import org.ow2.util.plan.deployer.api.FragmentDeploymentException;
import org.ow2.util.plan.deployer.api.FragmentUndeploymentException;
import org.ow2.util.plan.deployer.api.IDeploymentPlanDeployer;
import org.ow2.util.plan.fetcher.api.IResourceFetcher;
import org.ow2.util.plan.fetcher.api.IResourceFetcherFactoryManager;
import org.ow2.util.plan.monitor.api.IResourceMonitor;
import org.ow2.util.plan.reader.plan.IPlanReader;

/**
 * Deployer for the deployment plan deployables.
 * @author Mickaël LEDUQUE
 */
public class DeploymentPlanDeployer implements IDeploymentPlanDeployer {

    /**
     * The DeployerManager that will be used to deploy parts of the deployment
     * plans.
     */
    private IDeployerManager deployerManager = null;

    /**
     * The Deployment -> Deployable mappings.
     */
    private Map<Deployment, IDeployable> deploymentToDeployableMap =
                                    new HashMap<Deployment, IDeployable>();

    /**
     * The resource monitor.
     */
    private IResourceMonitor resourceMonitor = null;

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(DeploymentPlanDeployer.class);

    /**
     * The deployed deployment plans.
     */
    private Collection<DeploymentPlanDeployable> deploymentPlans = new ArrayList<DeploymentPlanDeployable>();

    /**
     * The PlanReader that will be used to parse the deployment plan.
     */
    private IPlanReader planReader;

    /**
     * The resource fetchers manager.
     */
    private IResourceFetcherFactoryManager fetcherFactoryManager = null;

    /**
     * {@inheritDoc}
     */
    public void deploy(final IDeployable<?> deployable) throws DeployerException {
        if (deployable == null) {
            logger.error("Null deployable");
            throw new DeployerException("Null deployable");
        }
        if (!(deployable instanceof DeploymentPlanDeployable)) {
            logger.error("Bad deployable type {0}", deployable.getClass());
            throw new DeployerException("Bad deployable type");
        }
        DeploymentPlanDeployable deploymentPlanDeployable = (DeploymentPlanDeployable) deployable;

        DeploymentPlan deploymentPlan = null;
        if (planReader == null) {
            logger.error("Deployment aborted - the PlanReader is null");
            throw new DeployerException("the PlanReader is null");
        }
        URL archiveUrl = null;
        try {
            archiveUrl = deploymentPlanDeployable.getArchive().getURL();
        } catch (ArchiveException e) {
            logger.error("Error while getting URL for archive {0} - Deployment aborted", deploymentPlanDeployable.getArchive());
            throw new DeployerException(e);
        }
        try {
            deploymentPlan = planReader.readPlan(new File(archiveUrl.getPath()));
        } catch (Exception e) {
            logger.error("Error while parsing archive {0} - Deployment aborted", archiveUrl);
            throw new DeployerException(e);
        }

        if (deploymentPlan == null) {
            /* Some strange error occured */
            logger.error("The deployable {0} resolved in null deployment plan - Deployment aborted", deployable);
            throw new DeployerException("null deployment plan structure");
        }
        deploymentPlanDeployable.setAttachedData(deploymentPlan);

        try {
            deployDeploymentPlans(deploymentPlan);
        } catch (DeploymentPlanDeploymentException e) {
            logger.error("Deployment plan deployment failed; exception: {0}", e);
            throw new DeployerException(e);
        }

        logger.debug("deployable {0} deployed", deployable);
        this.deploymentPlans.add(deploymentPlanDeployable);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isDeployed(final IDeployable<?> deployable) throws DeployerException {
        return this.deploymentPlans.contains(deployable);
    }

    /**
     * {@inheritDoc}
     */
    public boolean supports(final IDeployable<?> deployable) {
        if (deployable instanceof DeploymentPlanDeployableImpl) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void undeploy(final IDeployable<?> deployable) throws DeployerException {
        if (deployable == null) {
            logger.error("Null deployable");
            throw new DeployerException("Null deployable");
        }
        if (!(deployable instanceof DeploymentPlanDeployable)) {
            logger.error("Bad deployable type {0}", deployable.getClass());
            throw new DeployerException("Bad deployable type");
        }
        if (isDeployed(deployable)) {
            this.deploymentPlans.remove(deployable);
            DeploymentPlanDeployable deploymentPlanDeployable = (DeploymentPlanDeployable) deployable;
            DeploymentPlan deploymentPlan = deploymentPlanDeployable.getAttachedData();
            try {
                undeployDeploymentPlans(deploymentPlan);
            } catch (DeploymentPlanUndeploymentException e) {
                logger.error("Deployment plan undeployment failed; exception: {0}", e);
                throw new DeployerException(e);
            }

            logger.debug("deployable {0} undeployed", deployable);
        } else {
            logger.error("Cannot undeploy deployable {0} - it is not deployed", deployable);
            throw new DeployerException("Cannot undeploy deployable that is not deployed");
        }
    }

    /**
     * {@inheritDoc}
     */
    public void setPlanReader(final IPlanReader planReader) {
        this.planReader = planReader;
    }

    /**
     * {@inheritDoc}
     */
    public IPlanReader getPlanReader() {
        return this.planReader;
    }

    /**
     * Deploys all deployment plans object.
     * @param deploymentPlan the deployment plan.
     * @throws DeploymentPlanDeploymentException
     */
    private void deployDeploymentPlans(final DeploymentPlan deploymentPlan)
                                            throws DeploymentPlanDeploymentException {
        for (DeploymentPlan plans : deploymentPlan.getDeploymentPlans()) {
            deployDeploymentPlans(plans);
        }
        deployDeploymentPlan(deploymentPlan);
    }

    /**
     * Deploys a deployment plan object.
     * @param deploymentPlan the deployment plan.
     * @throws DeploymentPlanDeploymentException
     */
    private void deployDeploymentPlan(final DeploymentPlan deploymentPlan)
                                            throws DeploymentPlanDeploymentException {
        for (Deployment deployment : deploymentPlan.getDeployments()) {
            DeploymentPlanFragment fragment = (DeploymentPlanFragment) deployment;
            try {
                deployDeployment(fragment);
                fragment.setDeployed(true);
            } catch (FragmentDeploymentException deployException) {
                logger.error(
                        "Error while deploying part {0} of deployment plan {1} : {2}",
                        fragment, deploymentPlan, deployException);
                // One deployment has failed. Is the deployment plan atomic?
                if (deploymentPlan.isAtomic()) {
                    logger.error(
                            "The deployment plan {0} failed and is atomic; it must be undeployed",
                            deploymentPlan);
                    // try to undeploy everything
                    try {
                        undeployDeploymentPlans(deploymentPlan);
                    } catch (DeploymentPlanUndeploymentException undeployException) {
                        logger.error("The deployment plan was not correctly undeployed.");
                    }
                    throw new DeploymentPlanDeploymentException(deployException);
                }
            }
        }
    }

    /**
     * Undeploys all deployment plans object.
     * @param deploymentPlan the deployment plan.
     * @throws DeploymentPlanDeploymentException
     * @throws DeploymentPlanUndeploymentException
     */
    private void undeployDeploymentPlans(final DeploymentPlan deploymentPlan)
                                            throws DeploymentPlanUndeploymentException {
        for (DeploymentPlan plans : deploymentPlan.getDeploymentPlans()) {
            undeployDeploymentPlans(plans);
        }
        undeployDeploymentPlan(deploymentPlan);
    }

    private void undeployDeploymentPlan(final DeploymentPlan deploymentPlan) throws DeploymentPlanUndeploymentException {
        for (Deployment deployment : deploymentPlan.getDeployments()) {
            DeploymentPlanFragment fragment = (DeploymentPlanFragment) deployment;
            boolean partialFail = false;
            if (fragment.isDeployed()) {
                try {
                    undeployDeployment(fragment);
                    fragment.setDeployed(false);
                } catch (FragmentUndeploymentException e) {
                    logger.error("error undeploying deployment {0} of deployment plan {1}",
                           fragment, deploymentPlan);
                    /* Will continue to try undeploying other resources */
                    partialFail = true;
                }
            }
            /* We've done as much as we could, now we can fail */
            /* Note : we've lost the exception stack */
            if (partialFail == true) {
                throw new DeploymentPlanUndeploymentException();
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void deployDeployment(final Deployment deployment) throws FragmentDeploymentException {
        if (deployment == null) {
            throw new FragmentDeploymentException("Null deployment fragment.");
        }
        if (!(deployment instanceof DeploymentPlanFragment)) {
            throw new FragmentDeploymentException("Bad deployment instance : " + deployment);
        }
        DeploymentPlanFragment fragment = (DeploymentPlanFragment)deployment;

        Boolean reloadable = null;
        try {
            reloadable = DeploymentHelper.isReloadable(fragment);
        } catch (InvalidDeploymentException e) {
            logger.error("Invalid deployment found : {0} with id {1}", fragment);
            throw new FragmentDeploymentException(e);
        }

        File file = null;
        IResourceFetcher fetcher = null;
        try {
            fetcher = this.fetcherFactoryManager.getResourceFetcher(fragment);
            fetcher.resolve();
            file = fetcher.getResource();
        } catch (Exception e) {
            throw new FragmentDeploymentException(e);
        }

        /*
         * file is a *local* file containing the resource. We just need to
         * deploy it the usual way.
         */
        if (file == null) {
            throw new FragmentDeploymentException("Null file");
        }
        /* Create an archive */
        IArchive archive = ArchiveManager.getInstance().getArchive(file);
        if (archive == null) {
            throw new FragmentDeploymentException("Null archive for file " + file);
        }
        /* Create a deployable */
        IDeployable<?> deployable = null;
        try {
            deployable = DeployableHelper.getDeployable(archive);
        } catch (DeployableHelperException e) {
            throw new FragmentDeploymentException(e);
        }
        if (deployable == null) {
            throw new FragmentDeploymentException("Null deployable for archive" + archive);
        }

        // Check the start property to define if the bundle must be started or not
        if (deployable instanceof OSGiDeployable) {
            try {
                Boolean start = DeploymentHelper.isStart(fragment);
                ((OSGiDeployable) deployable).setStart(start);
            } catch (InvalidDeploymentException e) {
                logger.error("Invalid deployment found : {0} with id {1}", fragment);
                throw new FragmentDeploymentException(e);
            }
        }

        /* Deploy the deployable */
        try {
            this.deployerManager.deploy(deployable);
        } catch (DeployerException e) {
            logger.error("Exception while deploying deployable {0}: {1}", deployable, e.getMessage());
            throw new FragmentDeploymentException(e);
        } catch (UnsupportedDeployerException e) {
            logger.error("No deployer found for deployable {0}", deployable);
            throw new FragmentDeploymentException(e);
        }

        // Store the deployment->deployable
        deploymentToDeployableMap.put(fragment, deployable);
        if (reloadable) {
            // add this resource to the managed ones.
            this.resourceMonitor.addMonitoredResource(fetcher);
        }
    }

    /**
     * {@inheritDoc}
     */
    /* The only cases this will be used are : 1. when undeploying a deployment plan and 2. when
    redeploying a modified resource (by the monitor) */
    public void undeployDeployment(final Deployment deployment) throws FragmentUndeploymentException {
        if (deployment == null) {
            throw new FragmentUndeploymentException("Null deployment fragment.");
        }
        // find the deployable and undeploy it
        IDeployable deployable = deploymentToDeployableMap.get(deployment);
        if (deployable == null) {
            throw new FragmentUndeploymentException("Cannot undeploy Deployment " + deployment.toString());
        }
        deploymentToDeployableMap.remove(deployment);

        try {
            this.deployerManager.undeploy(deployable);
        } catch (Exception e) {
            logger.error(
                    "Could not undeploy resource {0} - trying to finish undeployment",
                    deployable.getArchive());
            throw new FragmentUndeploymentException(e);
        }

    }

    public void setFetcherFactoryManager(final IResourceFetcherFactoryManager factoryManager) {
        this.fetcherFactoryManager = factoryManager;
    }

    /**
     * {@inheritDoc}
     */
    public IResourceFetcherFactoryManager getFetcherFactoryManager() {
        return this.fetcherFactoryManager;
    }

    /**
     * {@inheritDoc}
     */
    public void setDeployerManager(final IDeployerManager deployerManager) {
        this.deployerManager = deployerManager;
    }

    /**
     * {@inheritDoc}
     */
    public IDeployerManager getDeployerManager() {
        return this.deployerManager;
    }

    /**
     * {@inheritDoc}
     */
    public Integer getDeploymentPlansCount() {
        return this.deploymentPlans.size();
    }

    /**
     * {@inheritDoc}
     */
    public List<String> getDeploymentPlansIds() {
        List<String> result = new ArrayList<String>();
        for (DeploymentPlanDeployable deploymentPlanDeployable : this.deploymentPlans) {
            DeploymentPlan deploymentPlan = deploymentPlanDeployable.getAttachedData();
            if (deploymentPlan.getId() == null) {
                result.add("<unnamed deployment>");
            } else {
                result.add(deploymentPlan.getId());
            }

        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public IResourceMonitor getResourceMonitor() {
        return this.resourceMonitor;
    }

    /**
     * {@inheritDoc}
     */
    public void setResourceMonitor(final IResourceMonitor resourceMonitor) {
        this.resourceMonitor = resourceMonitor;
    }

    /**
     * {@inheritDoc}
     */
    public DeploymentPlanDeployable getOwnerDeploymentPlanDeployable(final Deployment deployment) {
        for (DeploymentPlanDeployable deploymentPlanDeployable : this.deploymentPlans) {
            DeploymentPlan deploymentPlan = deploymentPlanDeployable.getAttachedData();
            if (deploymentPlan.getDeployments().contains(deployment)) {
                return deploymentPlanDeployable;
            }
        }
        return null;
    }
}
